import logging
from typing import List, Union, Optional

from ldap import UNWILLING_TO_PERFORM
from ldap.ldapobject import SimpleLDAPObject as LdapObject

from . import _ldaps, _groups, exceptions
from ._object_classes import User, Group, create_users
from ._query import object_exists, get, modify

logger = logging.getLogger(__name__)

NOT_DELEGATED = 1_048_576
EXPIRED_PASSWORD = 8388608
DONT_EXPIRE_PASSWORD = 65536
ENCRYPTED_TEXT_PWD_ALLOWED = 128
SMARTCARD_REQUIRED = 262_144
PASSWD_NOTREQD = 32
USER_BASE_QUERY = "(&(|(objectClass=user)(objectClass=person))(!(objectClass=computer)))"


@create_users
def _get(
    connection: LdapObject,
    base_query: Optional[str] = USER_BASE_QUERY,
    search_base: Optional[str] = None,
    search_scope: Optional[int] = None,
    **query_params,
) -> List[User]:
    return get(base_query, connection, search_base=search_base, search_scope=search_scope, **query_params,)


def _create(connection: LdapObject, distinguished_name: str, samaccountname: str, **extra_params) -> None:
    attr_dict = {
        "distinguishedName": bytes(distinguished_name, "utf-8"),
        "sAMAccountName": bytes(samaccountname, "utf-8"),
        "objectClass": [b"top", b"person", b"user"],
        "userAccountControl": b"514",  # disabled with no password set
    }
    attr_dict.update(_ldaps._encode_attributes(extra_params))
    return _ldaps._create(connection, distinguished_name, attr_dict)


@object_exists(_get, "users")
def _has_uac_flag(connection: LdapObject, distinguished_name: str, *, flag: int) -> bool:
    return bool(
        _get(
            connection, **{"distinguishedName": distinguished_name, "userAccountControl:1.2.840.113556.1.4.803:": flag}
        )
    )


@object_exists(_get, "users")
def _by_distinguished_name(connection: LdapObject, distinguished_name: str) -> User:
    return _get(connection, distinguishedName=distinguished_name)[0]


@object_exists(_get, "users")
def _delete(connection: LdapObject, distinguished_name: str) -> None:
    return _ldaps._delete(connection, distinguished_name)


@object_exists(_get, "users")
def _move(connection: LdapObject, distinguished_name: str, new_parent: str) -> None:
    return _ldaps._move(connection, distinguished_name, new_parent)


@object_exists(_get, "users")
def _group_memberships(connection: LdapObject, distinguished_name: str, nested: Optional[bool] = False) -> List[Group]:
    return _groups._get(
        connection,
        **{"member:1.2.840.113556.1.4.1941:": distinguished_name} if nested else {"member": distinguished_name},
    )


@object_exists(_groups._get, "groups")
def _not_members_of(
    connection: LdapObject, group_distinguished_names: List[str], search_base: Optional[str] = None
) -> List[User]:
    # Right now the build query params function doesn't support negative look-aheads
    # so this function sends the whole query as the base query. This is fine as we
    # don't need to support any extra search params for this operation anyway.
    query = "(&{base}{membership})".format(
        base=USER_BASE_QUERY,
        membership="".join([f"(!(memberOf={distinguished_name}))" for distinguished_name in group_distinguished_names]),
    )
    return _get(connection, base_query=query, search_base=search_base)


@object_exists(_groups._get, "groups")
def _members_of(
    connection: LdapObject, group_distinguished_names: List[str], search_base: Optional[str] = None
) -> List[User]:
    # Right now the build query params function doesn't support negative look-aheads
    # so this function sends the whole query as the base query. This is fine as we
    # don't need to support any extra search params for this operation anyway.
    query = "(&{base}{membership})".format(
        base=USER_BASE_QUERY,
        membership=f"(memberOf={group_distinguished_names[0]})"
        if len(group_distinguished_names) == 1
        else "(|{groups})".format(
            groups="".join([f"(memberOf={distinguished_name})" for distinguished_name in group_distinguished_names])
        ),
    )
    return _get(connection, base_query=query, search_base=search_base)


@object_exists(_get, "users")
def _modify(
    connection: LdapObject, distinguished_name: str, key: str, value: Union[int, str, bytes], operation: str
) -> None:
    return modify(connection, distinguished_name, key, value, operation)


def _enable(connection: LdapObject, distinguished_name: str) -> None:
    try:
        return _modify(connection, distinguished_name, "userAccountControl", b"512", "replace")
    except UNWILLING_TO_PERFORM as e:
        exception_data = _ldaps._process_ldap_exception(e)
        if exception_data.error == "0000052D":
            raise exceptions.OperationFailed(
                f"Failed to enable user with distinguishedName '{distinguished_name}' ."
                "The user may not have a password set."
            )
        raise e


def _set_password(connection: LdapObject, distinguished_name: str, new_password: str) -> None:
    encoded_password: bytes = f'"{new_password}"'.encode("utf-16-le")
    try:
        _modify(connection, distinguished_name, "unicodePwd", encoded_password, "replace")
    except UNWILLING_TO_PERFORM:
        raise exceptions.InvalidPassword(
            "Server has rejected the new password. This may be because it does not meet the domain password policy."
        )
