import logging
from typing import Union, List, Optional

from . import client, _users
from ._ldaps import LdapObject
from ._object_classes import User, Group

logger = logging.getLogger(__name__)


def test_credentials(*, address: str, port: Union[str, int], domain: str, username: str, password: str) -> None:
    """ Test an arbitrary set of user credentials against the
    Active Directory LDAP server. Useful to use after resetting a password
    to close the loop.

    :param address: The hostname or IP address of the domain controller
        to connect to.
    :param port: The listening LDAPS port (usually 636).
    :param domain: The domain FQDN.
    :param username: The username to bind with.
    :param password: The password to bind with.

    :raises exceptions.UserAccountLocked: if the supplied user account is locked.
    :raises exceptions.UserAccountExpired: if the supplied user account has expired.
    :raises exceptions.PasswordExpired: if the supplied user account has an expired password.
    :raises exceptions.PasswordResetRequired: if the password for the supplied user account must be reset.
    :raises exceptions.LogonNotPermitted: if the supplied user account is not allowed to logon at this time.
    :raises exceptions.InvalidCredentials: if authentication using the supplied credentials failed.
    :raises exceptions.ServerDown: if connectivity to the supplied address failed.

    Example:

        .. code-block:: python

            users.test_credentials(
                address='host',
                port='636',
                domain='my.domain.com',
                username='username',
                password='password'
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"testing credentials for username {username}")
    session = client.initialize(address=address, port=port, domain=domain, username=username, password=password)
    with session() as _:
        pass


def search(connection: LdapObject, search_base: Optional[str] = None, **search_params) -> List[User]:
    """ Search for an Active Directory user.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once automatically escaped,
        result in an invalid LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory user objects.

    To find the user with sAMAccountName 'test user ', call the function like so:

    .. code-block:: python

        users.search(connection, sAMAccountName="test user ")

    Wildcards are supported provided they are at the end of the search string. To find all users with a sAMAccountName
    starting with 'david' call the function like so:

    .. code-block:: python

        users.search(connection, sAMAccountName="david*")

    .. note:: If the wildcard character is not at the end of the search string, it will be escaped.
    """
    logger.info(f"searching for user with params {search_params}")
    return _users._get(connection, search_base=search_base, **search_params)


def get_all(connection: LdapObject, search_base: Optional[str] = None) -> List[User]:
    """ Get all Active Directory users.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.

    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory user objects.

    Example:

        .. code-block:: python

            users.get_all(connection)

    .. warning:: This operation can be expensive! Using the search function is recommended.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    logger.info("getting all users")
    return _users._get(connection, search_base=search_base)


def create(connection: LdapObject, *, distinguished_name: str, samaccountname: str, **extra_params) -> None:
    """ Create a new Active Directory user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguishedName for the new user.
    :param samaccountname: The sAMAccountName for the new user.
    :param extra_params: Keyword arguments to include in the new user LDAP attributes.

    :raises exceptions.ObjectExists: if the supplied user already exists.
    :raises exceptions.ParameterError:  if an invalid LDAP attribute is supplied.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.create(
                connection,
                distinguished_name='cn=my new user,cn=users,dc=test,dc=domain,dc=com',
                samaccountname='my.new.user',
                userPrincipalName='user@domain.com'
            )

    .. seealso:: :ref:`active_directory:Action Functions`, :ref:`active_directory:Keyword Argument Attributes`
    """
    logger.info(
        f"creating user with distinguished name: {distinguished_name}, "
        f"samaccountname: {samaccountname}, extra params: {extra_params}"
    )
    return _users._create(connection, distinguished_name, samaccountname, **extra_params)


def delete(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Delete a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.delete(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"deleting user with distinguished name {distinguished_name}")
    return _users._delete(connection, distinguished_name)


def move(connection: LdapObject, *, distinguished_name: str, new_parent: str) -> None:
    """ Move a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param new_parent: The distinguished name of the new OU or CN to put the user in.

    :raises exceptions.NoSuchObject: if the user or new parent does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.move(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
                new_parent='ou=staff,dc=test,dc=domain,dc=com'
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"moving user with distinguished name {distinguished_name} to location {new_parent}")
    return _users._move(connection, distinguished_name, new_parent)


def not_members_of(
    connection: LdapObject, *, group_distinguished_names: List[str], search_base: Optional[str] = None,
) -> List[User]:
    """Find all users who are not a member of any of the supplied groups.

    :param connection: An LDAPS connection.
    :param group_distinguished_names: A list of group distinguishedNames.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory user objects.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    logger.info(f"finding users that aren't a member of any of the following groups {group_distinguished_names}")
    return _users._not_members_of(connection, group_distinguished_names, search_base=search_base)


def members_of(
    connection: LdapObject, *, group_distinguished_names: List[str], search_base: Optional[str] = None,
) -> List[User]:
    """Find all users who are a member of any of the supplied groups.

    :param connection: An LDAPS connection.
    :param group_distinguished_names: A list of group distinguishedNames.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory user objects.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    logger.info(f"finding users that are a member of any of the following groups {group_distinguished_names}")
    return _users._members_of(connection, group_distinguished_names, search_base=search_base)


def group_memberships(
    connection: LdapObject, *, distinguished_name: str, nested: Optional[bool] = False
) -> List[Group]:
    """ List the group memberships of a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param nested: Optional. Defaults to False. If set to `True` the function will return both direct and
        nested group memberships.

    .. important:: Finding nested memberships can be very slow!

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory group objects.

    .. note:: This function does not find nested group memberships.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    logger.info(
        f"getting {'nested memberships' if nested else 'immediate memberships'} "
        f"for user with distinguished name {distinguished_name}"
    )
    return _users._group_memberships(connection, distinguished_name, nested=nested)


def is_locked(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account is locked.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account is locked. `False` if it is not locked.
    """
    return _users._by_distinguished_name(connection, distinguished_name=distinguished_name).is_locked


def is_enabled(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account is enabled. If the user's `userAccountControl` value **is not** one
    of `514`, `522`, `546`, it is considered **enabled**.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account is enabled. `False` if it is disabled.
    """
    return _users._by_distinguished_name(connection, distinguished_name=distinguished_name).is_enabled


def is_disabled(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account is enabled. If the user's `userAccountControl` value **is** one
    of `514`, `522`, `546`, it is considered **disabled**.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account is disabled. `False` if it is enabled.
    """
    return _users._by_distinguished_name(connection, distinguished_name=distinguished_name).is_disabled


def is_expired(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account is expired. This is established by converting the `accountExpires` attribute to
    a UTC timestamp, and comparing it to the UTC timestamp the moment the function is called.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account is expired. `False` if it is not expired.
    """
    return _users._by_distinguished_name(connection, distinguished_name=distinguished_name).is_expired


def password_has_expired(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account password has expired.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account password has expired. `False` if it has not.
    """
    return _users._has_uac_flag(connection, distinguished_name=distinguished_name, flag=_users.EXPIRED_PASSWORD)


def password_will_expire(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account has a password that will expire.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account password is set to expire. `False` if it is not.
    """
    return not _users._has_uac_flag(connection, distinguished_name=distinguished_name, flag=_users.DONT_EXPIRE_PASSWORD)


def password_not_required(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account can have a blank password.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account can have a blank password. `False` if it cannot.
    """
    return not _users._has_uac_flag(connection, distinguished_name=distinguished_name, flag=_users.PASSWD_NOTREQD)


def can_be_delegated(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account **has not** been flagged as 'sensitive and not for delegation'.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the account can be delegated. `False` if it cannot.
    """
    return not _users._has_uac_flag(connection, distinguished_name=distinguished_name, flag=_users.NOT_DELEGATED)


def has_reversible_password(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account password is stored using reversible encryption.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if the password is reversible. `False` if it is not.
    """
    return _users._has_uac_flag(
        connection, distinguished_name=distinguished_name, flag=_users.ENCRYPTED_TEXT_PWD_ALLOWED
    )


def smart_card_required(connection: LdapObject, *, distinguished_name: str) -> bool:
    """ Determine if the user account requires a smart card for interactive login.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: `True` if a smart card is required. `False` if it is not.
    """
    return _users._has_uac_flag(connection, distinguished_name=distinguished_name, flag=_users.SMARTCARD_REQUIRED)


def set_password(connection: LdapObject, *, distinguished_name: str, new_password: str) -> None:
    """ Set a new password for a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param new_password: The new password to set.

    :raises exceptions.InvalidPassword: if the domain rejects the new password.
    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.set_password(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
                new_password="N3wp45sw04d"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"setting password for user {distinguished_name}")
    return _users._set_password(connection, distinguished_name, new_password)


def force_password_change(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Force a user's password to be changed at next logon.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.force_password_change(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"forcing password change for user {distinguished_name}")
    return _users._modify(connection, distinguished_name, "pwdLastSet", "0", "replace")


def unlock(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Unlock a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.unlock(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"unlocking user {distinguished_name}")
    return _users._modify(connection, distinguished_name, "lockoutTime", "0", "replace")


def enable(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Enable a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.enable(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"enabling user {distinguished_name}")
    return _users._enable(connection, distinguished_name)


def disable(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Disable a user.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.

    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.disable(
                connection,
                distinguished_name='cn=my old user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"disabling user {distinguished_name}")
    return _users._modify(connection, distinguished_name, "userAccountControl", b"514", "replace")


def exists(connection: LdapObject, **search_params) -> bool:
    """ Check whether a user exists.

    :param connection: An LDAPS connection.
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once escaped, result in an invalid
        LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: Bool.

    Example:

        .. code-block:: python

            users.exists(connection, sAMAccountName="test user ")

    .. seealso:: :ref:`active_directory:Boolean Functions`, :ref:`active_directory:Keyword Argument Attributes`
    """
    logger.info(f"checking user exists with params {search_params}")
    return bool(search(connection, **search_params))


def set_attribute(connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]) -> None:
    """ Set a user LDAP attribute value. This function performs the LDAP modify_add
    operation. It is suitable for setting values for blank attributes.

    If the supplied attribute already has a value, exceptions.ModifyAttributeError
    will be raised. In this case modify_attribute should be used instead.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param name: The name of the LDAP attribute to modify.
    :param value: The value to set on the LDAP attribute.

    :raises exceptions.ModifyAttributeError: if there was an issue setting the new attribute value.
    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.set_attribute(
                connection,
                distinguished_name="cn=test user,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"setting attribute {name} to {value} on user {distinguished_name}")
    _users._modify(connection, distinguished_name, name, value, "add")


def replace_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Modify a user LDAP attribute value. This function performs the LDAP modify_replace
    operation.

    As suggested by the function name; if the supplied attribute already has a value, it will
    be replaced.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param name: The name of the LDAP attribute to modify.
    :param value: The value to replace the existing LDAP attribute value with.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the user does not exist.

    Example:

        .. code-block:: python

            users.replace_attribute(
                connection,
                distinguished_name="cn=test user,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my replaced value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"replacing attribute {name} to {value} on user {distinguished_name}")
    _users._modify(connection, distinguished_name, name, value, "replace")


def delete_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Remove a user LDAP attribute value. This function performs the LDAP modify_delete
    operation.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param name: The name of the LDAP attribute to modify.
    :param value: The attribute value to delete.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the user does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            users.delete_attribute(
                connection,
                distinguished_name="cn=test user,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="known existing value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"deleting attribute {name} value {value} on user {distinguished_name}")
    _users._modify(connection, distinguished_name, name, value, "delete")
